@@ -22,6 +22,7 @@
 
 #include "AsyncSocket.h"
 #include "HttpResponseData.h"
+#include "HttpContext.h"
 #include "HttpContextData.h"
 #include "Utilities.h"
 
@@ -30,7 +31,7 @@
 #include "WebSocket.h"
 #include "WebSocketContextData.h"
 
-#include "f2/function2.hpp"
+#include "MoveOnlyFunction.h"
 
 /* todo: tryWrite is missing currently, only send smaller segments with write */
 
@@ -86,14 +87,14 @@ private:
 #ifndef UWS_HTTPRESPONSE_NO_WRITEMARK
         if (!Super::getLoopData()->noMark) {
             /* We only expose major version */
-            writeHeader("uWebSockets", "18");
+            writeHeader("uWebSockets", "20");
         }
 #endif
     }
 
     /* Returns true on success, indicating that it might be feasible to write more data.
      * Will start timeout if stream reaches totalSize or write failure. */
-    bool internalEnd(std::string_view data, size_t totalSize, bool optional, bool allowContentLength = true) {
+    bool internalEnd(std::string_view data, uintmax_t totalSize, bool optional, bool allowContentLength = true, bool closeConnection = false) {
         /* Write status if not already done */
         writeStatus(HTTP_200_OK);
 
@@ -103,6 +104,22 @@ private:
         }
 
         HttpResponseData<SSL> *httpResponseData = getHttpResponseData();
+
+        /* In some cases, such as when refusing huge data we want to close the connection when drained */
+        if (closeConnection) {
+
+            /* HTTP 1.1 must send this back unless the client already sent it to us.
+             * It is a connection close when either of the two parties say so but the
+             * one party must tell the other one so.
+             *
+             * This check also serves to limit writing the header only once. */
+            if ((httpResponseData->state & HttpResponseData<SSL>::HTTP_CONNECTION_CLOSE) == 0) {
+                writeHeader("Connection", "close");
+            }
+
+            httpResponseData->state |= HttpResponseData<SSL>::HTTP_CONNECTION_CLOSE;
+        }
+
         if (httpResponseData->state & HttpResponseData<SSL>::HTTP_WRITE_CALLED) {
 
             /* We do not have tryWrite-like functionalities, so ignore optional in this path */
@@ -198,7 +215,7 @@ public:
             struct us_socket_context_t *webSocketContext) {
 
         /* Extract needed parameters from WebSocketContextData */
-        WebSocketContextData<SSL> *webSocketContextData = (WebSocketContextData<SSL> *) us_socket_context_ext(SSL, webSocketContext);
+        WebSocketContextData<SSL, UserData> *webSocketContextData = (WebSocketContextData<SSL, UserData> *) us_socket_context_ext(SSL, webSocketContext);
 
         /* Note: OpenSSL can be used here to speed this up somewhat */
         char secWebSocketAccept[29] = {};
@@ -214,67 +231,49 @@ public:
             writeHeader("Sec-WebSocket-Protocol", secWebSocketProtocol.substr(0, secWebSocketProtocol.find(',')));
         }
 
-        /* Negotiate compression, we may use a smaller compression window than we negotiate */
+        /* Negotiate compression */
         bool perMessageDeflate = false;
-        /* We are always allowed to share compressor, if perMessageDeflate */
-        int compressOptions = webSocketContextData->compression & SHARED_COMPRESSOR;
-        if (webSocketContextData->compression != DISABLED) {
-            if (secWebSocketExtensions.length()) {
-                /* We never support client context takeover (the client cannot compress with a sliding window). */
-                unsigned int wantedOptions = PERMESSAGE_DEFLATE | CLIENT_NO_CONTEXT_TAKEOVER;
+        CompressOptions compressOptions = CompressOptions::DISABLED;
+        if (secWebSocketExtensions.length() && webSocketContextData->compression != DISABLED) {
 
-                /* Shared compressor is the default */
-                if (webSocketContextData->compression == SHARED_COMPRESSOR) {
-                    /* Disable per-socket compressor */
-                    wantedOptions |= SERVER_NO_CONTEXT_TAKEOVER;
-                }
+            /* Make sure to map SHARED_DECOMPRESSOR to windowBits = 0, not 1  */
+            int wantedInflationWindow = 0;
+            if ((webSocketContextData->compression & CompressOptions::_DECOMPRESSOR_MASK) != CompressOptions::SHARED_DECOMPRESSOR) {
+                wantedInflationWindow = (webSocketContextData->compression & CompressOptions::_DECOMPRESSOR_MASK) >> 8;
+            }
 
-                /* isServer = true */
-                ExtensionsNegotiator<true> extensionsNegotiator(wantedOptions);
-                extensionsNegotiator.readOffer(secWebSocketExtensions);
+            /* Map from selected compressor (this automatically maps SHARED_COMPRESSOR to windowBits 0, not 1) */
+            int wantedCompressionWindow = (webSocketContextData->compression & CompressOptions::_COMPRESSOR_MASK) >> 4;
 
-                /* Todo: remove these mid string copies */
-                std::string offer = extensionsNegotiator.generateOffer();
-                if (offer.length()) {
+            auto [negCompression, negCompressionWindow, negInflationWindow, negResponse] =
+            negotiateCompression(true, wantedCompressionWindow, wantedInflationWindow,
+                                        secWebSocketExtensions);
 
-                    /* Todo: this is a quick fix that should be properly moved to ExtensionsNegotiator */
-                    if (webSocketContextData->compression & DEDICATED_COMPRESSOR &&
-                        webSocketContextData->compression != DEDICATED_COMPRESSOR_256KB) {
-                            /* 3kb, 4kb is 9, 256 is 15 (default) */
-                            int maxServerWindowBits = 9;
-                            switch (webSocketContextData->compression) {
-                                case DEDICATED_COMPRESSOR_8KB:
-                                maxServerWindowBits = 10;
-                                break;
-                                case DEDICATED_COMPRESSOR_16KB:
-                                maxServerWindowBits = 11;
-                                break;
-                                case DEDICATED_COMPRESSOR_32KB:
-                                maxServerWindowBits = 12;
-                                break;
-                                case DEDICATED_COMPRESSOR_64KB:
-                                maxServerWindowBits = 13;
-                                break;
-                                case DEDICATED_COMPRESSOR_128KB:
-                                maxServerWindowBits = 14;
-                                break;
-                            }
-                            offer += "; server_max_window_bits=";
-                            offer += std::to_string(maxServerWindowBits);
+            if (negCompression) {
+                perMessageDeflate = true;
+
+                /* Map from negotiated windowBits to compressor and decompressor */
+                if (negCompressionWindow == 0) {
+                    compressOptions = CompressOptions::SHARED_COMPRESSOR;
+                } else {
+                    compressOptions = (CompressOptions) ((uint32_t) (negCompressionWindow << 4)
+                                                        | (uint32_t) (negCompressionWindow - 7));
+
+                    /* If we are dedicated and have the 3kb then correct any 4kb to 3kb,
+                     * (they both share the windowBits = 9) */
+                    if (webSocketContextData->compression & DEDICATED_COMPRESSOR_3KB) {
+                        compressOptions = DEDICATED_COMPRESSOR_3KB;
                     }
-
-                    writeHeader("Sec-WebSocket-Extensions", offer);
                 }
 
-                /* Did we negotiate permessage-deflate? */
-                if (extensionsNegotiator.getNegotiatedOptions() & PERMESSAGE_DEFLATE) {
-                    perMessageDeflate = true;
+                /* Here we modify the above compression with negotiated decompressor */
+                if (negInflationWindow == 0) {
+                    compressOptions = CompressOptions(compressOptions | CompressOptions::SHARED_DECOMPRESSOR);
+                } else {
+                    compressOptions = CompressOptions(compressOptions | (negInflationWindow << 8));
                 }
 
-                /* Is the server allowed to compress with a sliding window? */
-                if (!(extensionsNegotiator.getNegotiatedOptions() & SERVER_NO_CONTEXT_TAKEOVER)) {
-                    compressOptions = webSocketContextData->compression;
-                }
+                writeHeader("Sec-WebSocket-Extensions", negResponse);
             }
         }
 
@@ -284,7 +283,7 @@ public:
         HttpContext<SSL> *httpContext = (HttpContext<SSL> *) us_socket_context(SSL, (struct us_socket_t *) this);
 
         /* Move any backpressure out of HttpResponse */
-        std::string backpressure(std::move(((AsyncSocketData<SSL> *) getHttpResponseData())->buffer));
+        BackPressure backpressure(std::move(((AsyncSocketData<SSL> *) getHttpResponseData())->buffer));
 
         /* Destroy HttpResponseData */
         getHttpResponseData()->~HttpResponseData();
@@ -293,12 +292,12 @@ public:
         bool wasCorked = Super::isCorked();
 
         /* Adopting a socket invalidates it, do not rely on it directly to carry any data */
-        WebSocket<SSL, true> *webSocket = (WebSocket<SSL, true> *) us_socket_context_adopt_socket(SSL,
+        WebSocket<SSL, true, UserData> *webSocket = (WebSocket<SSL, true, UserData> *) us_socket_context_adopt_socket(SSL,
                     (us_socket_context_t *) webSocketContext, (us_socket_t *) this, sizeof(WebSocketData) + sizeof(UserData));
 
         /* For whatever reason we were corked, update cork to the new socket */
         if (wasCorked) {
-            webSocket->AsyncSocket<SSL>::cork();
+            webSocket->AsyncSocket<SSL>::corkUnchecked();
         }
 
         /* Initialize websocket with any moved backpressure intact */
@@ -312,7 +311,7 @@ public:
         }
 
         /* Arm idleTimeout */
-        us_socket_timeout(SSL, (us_socket_t *) webSocket, webSocketContextData->idleTimeout);
+        us_socket_timeout(SSL, (us_socket_t *) webSocket, webSocketContextData->idleTimeoutComponents.first);
 
         /* Move construct the UserData right before calling open handler */
         new (webSocket->getUserData()) UserData(std::move(userData));
@@ -331,6 +330,19 @@ public:
     using Super::getRemoteAddressAsText;
     using Super::getNativeHandle;
 
+    /* Throttle reads and writes */
+    HttpResponse *pause() {
+        Super::pause();
+        Super::timeout(0);
+        return this;
+    }
+
+    HttpResponse *resume() {
+        Super::resume();
+        Super::timeout(HTTP_TIMEOUT_S);
+        return this;
+    }
+
     /* Note: Headers are not checked in regards to timeout.
      * We only check when you actively push data or end the request */
 
@@ -371,6 +383,8 @@ public:
 
     /* Write an HTTP header with unsigned int value */
     HttpResponse *writeHeader(std::string_view key, uint64_t value) {
+        writeStatus(HTTP_200_OK);
+
         Super::write(key.data(), (int) key.length());
         Super::write(": ", 2);
         writeUnsigned64(value);
@@ -378,14 +392,23 @@ public:
         return this;
     }
 
+    /* End without a body (no content-length) or end with a spoofed content-length. */
+    void endWithoutBody(std::optional<size_t> reportedContentLength = std::nullopt, bool closeConnection = false) {
+        if (reportedContentLength.has_value()) {
+            //internalEnd({nullptr, 0}, reportedContentLength.value(), false, true, closeConnection);
+        } else {
+            internalEnd({nullptr, 0}, 0, false, false, closeConnection);
+        }
+    }
+
     /* End the response with an optional data chunk. Always starts a timeout. */
-    void end(std::string_view data = {}) {
-        internalEnd(data, data.length(), false);
+    void end(std::string_view data = {}, bool closeConnection = false) {
+        internalEnd(data, data.length(), false, true, closeConnection);
     }
 
     /* Try and end the response. Returns [true, true] on success.
      * Starts a timeout in some cases. Returns [ok, hasResponded] */
-    std::pair<bool, bool> tryEnd(std::string_view data, size_t totalSize = 0) {
+    std::pair<bool, bool> tryEnd(std::string_view data, uintmax_t totalSize = 0) {
         return {internalEnd(data, totalSize, true), hasResponded()};
     }
 
@@ -423,7 +446,7 @@ public:
     }
 
     /* Get the current byte write offset for this Http response */
-    size_t getWriteOffset() {
+    uintmax_t getWriteOffset() {
         HttpResponseData<SSL> *httpResponseData = getHttpResponseData();
 
         return httpResponseData->offset;
@@ -437,7 +460,7 @@ public:
     }
 
     /* Corks the response if possible. Leaves already corked socket be. */
-    HttpResponse *cork(fu2::unique_function<void()> &&handler) {
+    HttpResponse *cork(MoveOnlyFunction<void()> &&handler) {
         if (!Super::isCorked() && Super::canCork()) {
             Super::cork();
             handler();
@@ -458,7 +481,7 @@ public:
     }
 
     /* Attach handler for writable HTTP response */
-    HttpResponse *onWritable(fu2::unique_function<bool(size_t)> &&handler) {
+    HttpResponse *onWritable(MoveOnlyFunction<bool(uintmax_t)> &&handler) {
         HttpResponseData<SSL> *httpResponseData = getHttpResponseData();
 
         httpResponseData->onWritable = std::move(handler);
@@ -466,7 +489,7 @@ public:
     }
 
     /* Attach handler for aborted HTTP request */
-    HttpResponse *onAborted(fu2::unique_function<void()> &&handler) {
+    HttpResponse *onAborted(MoveOnlyFunction<void()> &&handler) {
         HttpResponseData<SSL> *httpResponseData = getHttpResponseData();
 
         httpResponseData->onAborted = std::move(handler);
@@ -474,7 +497,7 @@ public:
     }
 
     /* Attach a read handler for data sent. Will be called with FIN set true if last segment. */
-    void onData(fu2::unique_function<void(std::string_view, bool)> &&handler) {
+    void onData(MoveOnlyFunction<void(std::string_view, bool)> &&handler) {
         HttpResponseData<SSL> *data = getHttpResponseData();
         data->inStream = std::move(handler);
     }
